/*
 * Created by Julian Mutter on 7/10/18 3:58 PM
 * Copyright (c) 2018. All rights reserved.
 * Last modified 7/10/18 3:54 PM
 *
 */

package de.frajul.endlessroll.entities.collision;

import de.frajul.endlessroll.data.Vector;
import de.frajul.endlessroll.entities.Entity;
import de.frajul.endlessroll.entities.Player;
import de.frajul.endlessroll.entities.collision.collisionData.EntityCollisionData;
import de.frajul.endlessroll.entities.collision.geometry.Circle;
import de.frajul.endlessroll.entities.collision.geometry.Geometry;
import de.frajul.endlessroll.entities.collision.geometry.Quad;
import de.frajul.endlessroll.entities.collision.geometry.Triangle;
import de.frajul.endlessroll.entities.tileLists.Tile;

/**
 * Created by Julian on 01.12.2015.
 */
public class CollisionDetector {

    private CircleTriangleCollisionDetector triangleDetector;

    public CollisionDetector() {
        triangleDetector = new CircleTriangleCollisionDetector();
    }

    public boolean isCollision(Geometry g1, Geometry g2) {
        if (g1 instanceof Triangle) {
            if (g2 instanceof Triangle)
                return isTriangleTriangleCollision((Triangle) g1, (Triangle) g2);
            if (g2 instanceof Quad)
                return isQuadTriangleCollision((Quad) g2, (Triangle) g1);
            if (g2 instanceof Circle)
                return isCircleTriangleCollision((Circle) g2, (Triangle) g1);
        }
        if (g1 instanceof Quad) {
            if (g2 instanceof Triangle)
                return isQuadTriangleCollision((Quad) g1, (Triangle) g2);
            if (g2 instanceof Quad)
                return isQuadQuadCollision((Quad) g1, (Quad) g2);
            if (g2 instanceof Circle)
                return isCircleQuadCollision((Circle) g2, (Quad) g1);
        }
        if (g1 instanceof Circle) {
            if (g2 instanceof Triangle)
                return isCircleTriangleCollision((Circle) g1, (Triangle) g2);
            if (g2 instanceof Quad)
                return isCircleQuadCollision((Circle) g1, (Quad) g2);
            if (g2 instanceof Circle)
                return isCircleCircleCollision((Circle) g1, (Circle) g2);
        }
        return false;
    }

    public boolean isPointInQuad(Vector point, Quad quad) {
        if (point.getX() >= quad.getLeftEdge() && point.getX() <= quad.getRightEdge()) {
            return point.getY() >= quad.getBottomEdge() && point.getY() <= quad.getTopEdge();
        }
        return false;
    }

    public boolean isCircleCircleCollision(Circle c1, Circle c2) {
        float distance = Math.abs(c1.getPosition().vectorTo(c2.getPosition()).length());
        float radiusSum = c1.getRadius() + c2.getRadius();
        return distance < radiusSum;
    }

    public boolean isCircleQuadCollision(Circle circle, Quad quad) {
        Vector distance = circle.getPosition().vectorTo(quad.getPosition());
        distance.x = Math.abs(distance.x);
        distance.y = Math.abs(distance.y);

        if (distance.x > circle.getRadius() + quad.getWidth() / 2)
            return false;
        if (distance.y > circle.getRadius() + quad.getHeight() / 2)
            return false;

        if (distance.x <= quad.getWidth() / 2)
            return true;
        if (distance.y <= quad.getHeight() / 2)
            return true;

        float cornerDistance_sq = (distance.x - quad.getWidth() / 2) * (distance.x - quad
                .getWidth() / 2) + (distance.y - quad.getHeight() / 2) * (distance.y - quad
                .getHeight() / 2);
        return cornerDistance_sq <= circle.getRadius() * circle.getRadius();
    }

    public boolean isCircleTriangleCollision(Circle circle, Triangle triangle) {
        return triangleDetector.isCircleTriangleCollision(circle, triangle);
    }

    public boolean isQuadQuadCollision(Quad q1, Quad q2) {
        float xDistance = Math.abs(q2.getPosition().x - q1.getPosition().x);
        float yDistance = Math.abs(q2.getPosition().y - q1.getPosition().y);
        if (xDistance >= q1.getWidth() / 2 + q2.getWidth() / 2)
            return false;
        return !(yDistance >= q1.getHeight() / 2 + q2.getHeight() / 2);
    }

    public boolean isTriangleTriangleCollision(Triangle triangle1, Triangle triangle2) {
        boolean triangle1Vertex1InTriangle2 = isVertexInTriangle(triangle1.getBottomRightVertex(), triangle2);
        if (triangle1Vertex1InTriangle2)
            return true;
        boolean triangle1Vertex2InTriangle2 = isVertexInTriangle(triangle1.getTopRightVertex(), triangle2);
        if (triangle1Vertex2InTriangle2)
            return true;
        boolean triangle1Vertex3InTriangle2 = isVertexInTriangle(triangle1.getBottomLeftVertex(), triangle2);
        if (triangle1Vertex3InTriangle2)
            return true;
        boolean triangle2Vertex1InTriangle1 = isVertexInTriangle(triangle2.getBottomRightVertex(), triangle1);
        if (triangle2Vertex1InTriangle1)
            return true;
        boolean triangle2Vertex2InTriangle1 = isVertexInTriangle(triangle2.getTopRightVertex(), triangle1);
        if (triangle2Vertex2InTriangle1)
            return true;
        boolean triangle2Vertex3InTriangle1 = isVertexInTriangle(triangle2.getBottomLeftVertex(), triangle1);
        return triangle2Vertex3InTriangle1;
    }

    public boolean isQuadTriangleCollision(Quad quad, Triangle triangle) {
        boolean quadVertex1InTriangle = isVertexInTriangle(
                new Vector(quad.getLeftEdge(), quad.getTopEdge()), triangle);
        if (quadVertex1InTriangle)
            return true;
        boolean quadVertex2InTriangle = isVertexInTriangle(
                new Vector(quad.getRightEdge(), quad.getTopEdge()), triangle);
        if (quadVertex2InTriangle)
            return true;
        boolean quadVertex3InTriangle = isVertexInTriangle(
                new Vector(quad.getRightEdge(), quad.getBottomEdge()), triangle);
        if (quadVertex3InTriangle)
            return true;
        boolean quadVertex4InTriangle = isVertexInTriangle(
                new Vector(quad.getLeftEdge(), quad.getBottomEdge()), triangle);
        if (quadVertex4InTriangle)
            return true;
        boolean triangleVertex1InQuad = isPointInQuad(triangle.getBottomRightVertex(), quad);
        if (triangleVertex1InQuad)
            return true;
        boolean triangleVertex2InQuad = isPointInQuad(triangle.getTopRightVertex(), quad);
        if (triangleVertex2InQuad)
            return true;
        boolean triangleVertex3InQuad = isPointInQuad(triangle.getBottomLeftVertex(), quad);
        if (triangleVertex3InQuad)
            return true;
        return triangle.getTopEdge() >= quad.getTopEdge() && triangle.getBottomEdge() <= quad.getBottomEdge() && ((triangle.getLeftEdge() <= quad.getRightEdge() && triangle.getLeftEdge() >= quad.getLeftEdge()) || (triangle.getRightEdge() <= quad.getRightEdge() && triangle.getRightEdge() >= quad.getLeftEdge()));
    }

    private boolean isVertexInTriangle(Vector vertex, Triangle triangle) {
        if (!isPointInQuad(vertex, triangle))
            return false;
        float gradient = triangle.getGradient();
        float maxVertexY = vertex.getX() * gradient + triangle.getBottomEdge();
        return vertex.getY() <= maxVertexY;
    }

    public EntityCollisionData playerTerrainCollision(Player player, Tile terrainTile) {
        Circle circle = new Circle(player);
        if (circle.getTopEdge() < terrainTile.getTopEdge()) {
            float xDistance = Math.abs(circle.getPosition().x - terrainTile.getPosition().x);
            if (xDistance < circle.getRadius() + terrainTile.getWidth() / 2)
                return new EntityCollisionData(terrainTile, Edge.LEFT);
        }

        if (isCircleQuadCollision(circle, terrainTile)) {
            Edge edge = circleQuadCollisionEdge(circle, player.getMovement(), terrainTile,
                    terrainTile.getMovement());
            return new EntityCollisionData(terrainTile, edge);
        }
        return new EntityCollisionData(null, null);
    }

    public EntityCollisionData playerCeilingCollision(Player player, Tile ceilingTile) {
        Circle circle = new Circle(player);
        if (circle.getBottomEdge() > ceilingTile.getBottomEdge()) {
            float xDistance = Math.abs(circle.getPosition().x - ceilingTile.getPosition().x);
            if (xDistance < circle.getRadius() + ceilingTile.getWidth() / 2)
                return new EntityCollisionData(ceilingTile, Edge.LEFT);
        }

        if (isCircleQuadCollision(circle, ceilingTile)) {
            Edge edge = circleQuadCollisionEdge(circle, player.getMovement(), ceilingTile,
                    ceilingTile.getMovement());
            return new EntityCollisionData(ceilingTile, edge);
        }
        return new EntityCollisionData(null, null);
    }

    public EntityCollisionData playerEntityCollision(Player player, Entity entity) {
        Circle circle = new Circle(player);
        Quad quad = entity;
        if (isCircleQuadCollision(circle, quad)) {
            Edge edge = circleQuadCollisionEdge(circle, player.getMovement(), quad,
                    entity.getMovement());
            return new EntityCollisionData(entity, edge);
        }
        return new EntityCollisionData(null, null);
    }

    private Edge circleQuadCollisionEdge(Circle circle, Vector circleMovement, Quad quad, Vector quadMovement) {
        boolean circleFullyInsideQuad = false;
        if (circle.getPosition().getX() + circle.getRadius() <= quad.getRightEdge() && circle
                .getPosition().getX() - circle.getRadius() >= quad.getLeftEdge()) {
            if (circle.getPosition().getY() + circle.getRadius() <= quad.getTopEdge() && circle
                    .getPosition().getY() - circle.getRadius() >= quad.getBottomEdge()) {
                circleFullyInsideQuad = true;
            }
        }
        if (circleFullyInsideQuad)
            return circleCenterInQuadCollisionEdge(circle, circleMovement, quad, quadMovement);
        return circlePartlyOutsideQuadCollisionEdge(circle, circleMovement, quad, quadMovement);
    }

    private Edge circlePartlyOutsideQuadCollisionEdge(Circle circle, Vector circleMovement, Quad quad, Vector quadMovement) {
        Vector relativeMovement = new Vector(circleMovement).translate(quadMovement.negate());

        Edge edge = Edge.NONE;
        if (circle.getPosition().getY() + circle.getRadius() >= quad.getTopEdge())
            edge = Edge.TOP;
        else if (circle.getPosition().getY() - circle.getRadius() <= quad.getBottomEdge())
            edge = Edge.BOTTOM;
        if (circle.getPosition().getX() - circle.getRadius() <= quad.getLeftEdge()) {
            if (edge != Edge.NONE) {
                if (edge == Edge.TOP) {
                    float m = -1;
                    float t = quad.getTopEdge() - m * quad.getLeftEdge();
                    if (circle.getPosition().getY() >= m * circle.getPosition().getX() + t)
                        edge = Edge.TOP;
                    else
                        edge = Edge.LEFT;
                } else if (edge == Edge.BOTTOM) {
                    float m = 1;
                    float t = quad.getBottomEdge() - m * quad.getLeftEdge();
                    if (circle.getPosition().getY() <= m * circle.getPosition().getX() + t)
                        edge = Edge.BOTTOM;
                    else
                        edge = Edge.LEFT;
                }
            } else
                edge = Edge.LEFT;
        } else if (circle.getPosition().getX() + circle.getRadius() >= quad.getRightEdge()) {
            if (edge != Edge.NONE) {
                if (edge == Edge.TOP) {
                    //This movement check is because if player falls with too much speed on topLeft Vertex, sometimes RIGHT-Edge is returned
                    if (relativeMovement.getY() < 0) {
                        return Edge.TOP;
                    } else {
                        float m = 1;
                        float t = quad.getTopEdge() - m * quad.getRightEdge();
                        if (circle.getPosition().getY() >= m * circle.getPosition().getX() + t)
                            edge = Edge.TOP;
                        else
                            edge = Edge.RIGHT;
                    }
                } else if (edge == Edge.BOTTOM) {
                    float m = -1;
                    float t = quad.getBottomEdge() - m * quad.getRightEdge();
                    if (circle.getPosition().getY() <= m * circle.getPosition().getX() + t)
                        edge = Edge.BOTTOM;
                    else
                        edge = Edge.RIGHT;
                }
            } else
                edge = Edge.RIGHT;
        }
        return edge;
    }

    private Edge circleCenterInQuadCollisionEdge(Circle circle, Vector circleMovement, Quad quad, Vector quadMovement) {
        Vector relativeMovement = new Vector(circleMovement).translate(quadMovement.negate());

        float m = relativeMovement.getY() / relativeMovement.getX();
        float t = circle.getPosition().getY() - m * circle.getPosition().getX();

        float yAtLeftEdge = m * quad.getLeftEdge() + t;
        boolean touchingLeftEdge = yAtLeftEdge < quad.getTopEdge() && yAtLeftEdge > quad
                .getBottomEdge();
        float yAtRightEdge = m * quad.getRightEdge() + t;
        boolean touchingRightEdge = yAtRightEdge < quad.getTopEdge() && yAtRightEdge > quad
                .getBottomEdge();

        float xAtTopEdge = (quad.getTopEdge() - t) / m;
        boolean touchingTopEdge = xAtTopEdge >= quad.getLeftEdge() && xAtTopEdge <= quad
                .getRightEdge();
        float xAtBottomEdge = (quad.getBottomEdge() - t) / m;
        boolean touchingBottomEdge = xAtBottomEdge >= quad.getLeftEdge() && xAtBottomEdge <= quad
                .getRightEdge();

        if (relativeMovement.getX() == 0) {
            if (circle.getPosition().getX() - circle.getRadius() <= quad.getRightEdge() && circle
                    .getPosition().getX() + circle.getRadius() >= quad.getLeftEdge()) {
                touchingTopEdge = true;
                touchingBottomEdge = true;
            }
        }

        // CORNERS BELONGING TO TOP / BOTTOM EDGES
        if (touchingLeftEdge) {
            if (touchingRightEdge) {
                if (circle.getPosition().getX() <= quad.getPosition().getX())
                    return Edge.LEFT;
                return Edge.RIGHT;
            } else if (touchingTopEdge) {
                // Normale through vertex
                float mn = -1 / m;
                float tn = quad.getTopEdge() - mn * quad.getLeftEdge();
                if (circle.getPosition().getY() < mn * circle.getPosition().getX() + tn)
                    return Edge.LEFT;
                return Edge.TOP;
            } else if (touchingBottomEdge) {
                // Normale through vertex
                float mn = -1 / m;
                float tn = quad.getBottomEdge() - mn * quad.getLeftEdge();
                if (circle.getPosition().getY() > mn * circle.getPosition().getX() + tn)
                    return Edge.LEFT;
                return Edge.BOTTOM;
            } else {
                return Edge.LEFT;
            }
        } else if (touchingRightEdge) {
            if (touchingTopEdge) {
                // Normale through vertex
                float mn = -1 / m;
                float tn = quad.getTopEdge() - mn * quad.getRightEdge();
                if (circle.getPosition().getY() < mn * circle.getPosition().getX() + tn)
                    return Edge.RIGHT;
                return Edge.TOP;
            } else if (touchingBottomEdge) {
                // Normale through vertex
                float mn = -1 / m;
                float tn = quad.getBottomEdge() - mn * quad.getRightEdge();
                if (circle.getPosition().getY() > mn * circle.getPosition().getX() + tn)
                    return Edge.RIGHT;
                return Edge.BOTTOM;
            } else {
                return Edge.RIGHT;
            }
        } else if (touchingTopEdge) {
            if (touchingBottomEdge) {
                if (circle.getPosition().getY() >= quad.getPosition().getY())
                    return Edge.TOP;
                return Edge.BOTTOM;
            } else {
                return Edge.TOP;
            }
        } else if (touchingBottomEdge) {
            return Edge.BOTTOM;
        }
        return Edge.NONE;
    }
}
