/*
 * Created by Julian Mutter on 7/10/18 3:58 PM
 * Copyright (c) 2018. All rights reserved.
 * Last modified 7/10/18 3:54 PM
 *
 */

package de.frajul.endlessroll.main;

import android.app.Activity;
import android.app.ActivityManager;
import android.content.Context;
import android.content.pm.ConfigurationInfo;
import android.graphics.Typeface;
import android.os.Bundle;
import android.view.KeyEvent;
import android.view.Window;
import android.view.WindowManager;
import android.widget.RelativeLayout;

import com.google.android.gms.ads.MobileAds;

import java.util.ArrayList;
import java.util.List;

import de.frajul.endlessroll.entities.shapes.PlayerShape;
import de.frajul.endlessroll.entities.shapes.Task;
import de.frajul.endlessroll.entities.tools.ToolType;
import de.frajul.endlessroll.levels.Level;
import de.frajul.endlessroll.levels.LevelManager;
import de.frajul.endlessroll.levels.LevelPack;
import de.frajul.endlessroll.main.screens.GLTestScreen;
import de.frajul.endlessroll.main.screens.GameScreen;
import de.frajul.endlessroll.main.screens.LevelsScreen;
import de.frajul.endlessroll.main.screens.PlayerShapeShopScreen;
import de.frajul.endlessroll.main.screens.PreStartScreen;
import de.frajul.endlessroll.main.screens.Screen;
import de.frajul.endlessroll.main.screens.ScreenFlipper;
import de.frajul.endlessroll.main.screens.SettingsScreen;
import de.frajul.endlessroll.main.screens.StartScreen;
import de.frajul.endlessroll.main.screens.ToolShopScreen;
import de.frajul.endlessroll.main.screens.WorldsScreen;
import de.frajul.endlessroll.main.tutorial.BreakPoint;
import de.frajul.endlessroll.main.tutorial.TutorialManager;
import de.frajul.endlessroll.main.tutorial.TutorialView;
import de.frajul.endlessroll.rendering.renderer.GameRenderer;
import de.frajul.endlessroll.sounds.SoundManager;
import de.frajul.endlessroll.sqlDatabase.MyDatabase;
import de.frajul.endlessroll.user.User;
import de.frajul.endlessroll.views.LevelupMessage;
import de.frajul.endlessroll.views.TaskCompletedMessage;

/**
 * Created by Julian on 06.02.2016.
 */
public class GameActivity extends Activity implements ExceptionHandler, User.LvUpListener {

    public final static boolean IS_DEVELOPER = false;

    private DataStorageHandler dataStorageHandler;
    private LevelManager levelManager;
    private SoundManager soundManager;
    private User user;
    private Typeface typeface;
    private TutorialManager tutorialManager;

    private List<PlayerShape> alreadyUnlockedPlayerShapes = new ArrayList<>();

    private MyGlSurfaceView glSurfaceView;
    private ScreenFlipper flipper;
    private LevelsScreen levelsScreen;
    private GameScreen gameScreen;

    private LevelupMessage levelupMessage;
    private TaskCompletedMessage taskCompletedMessage;
    private TutorialView tutorialView;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        try {
            GameLog.d("OnCreate");
            super.onCreate(savedInstanceState);
            super.requestWindowFeature(Window.FEATURE_NO_TITLE);
            super.getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN,
                    WindowManager.LayoutParams.FLAG_FULLSCREEN);

            MobileAds.initialize(this, "ca-app-pub-7069685539417779~2685294395");
            if (!hasGLES20())
                throw new Exception("OpenGL ES 2.0 not supported");

            dataStorageHandler = new DataStorageHandler(this);
            dataStorageHandler.getDatabase().open();
            dataStorageHandler.getDatabase().readToolData();
            dataStorageHandler.getDatabase().close();

            user = dataStorageHandler.readUserData(this);

            soundManager = new SoundManager(this);
            soundManager.setSoundsMuted(dataStorageHandler.readIsSoundsMuted());
            soundManager.setMusicMuted(dataStorageHandler.readIsMusicMuted());

            levelManager = new LevelManager(this, dataStorageHandler);
            tutorialManager = new TutorialManager(this);
            tutorialManager.getToolShopTutorial()
                    .setFinished(dataStorageHandler.readToolShopTutorialFinished());

            checkForAlreadyUnlockedPlayerShapes();

            this.glSurfaceView = new MyGlSurfaceView(this, new GameRenderer(this));
            typeface = Typeface.createFromAsset(getAssets(), "fontBaron.ttf");

            PreStartScreen preStartScreen = new PreStartScreen(this);
            StartScreen startScreen = new StartScreen(this, glSurfaceView);
            GLTestScreen glTestScreen = new GLTestScreen(this, glSurfaceView);
            WorldsScreen worldsScreen = new WorldsScreen(this);
            levelsScreen = new LevelsScreen(this);
            ToolShopScreen toolShopScreen = new ToolShopScreen(this);
            PlayerShapeShopScreen shapeShopScreen = new PlayerShapeShopScreen(this);
            gameScreen = new GameScreen(this, glSurfaceView);
            SettingsScreen settingsScreen = new SettingsScreen(this);

            levelupMessage = new LevelupMessage(this);
            taskCompletedMessage = new TaskCompletedMessage(this);
            tutorialView = new TutorialView(this);

            flipper = new ScreenFlipper(this, preStartScreen, startScreen, worldsScreen,
                    levelsScreen, gameScreen, toolShopScreen, glTestScreen, settingsScreen,
                    shapeShopScreen);
            RelativeLayout relativeLayout = new RelativeLayout(this);
            relativeLayout.addView(glSurfaceView);
            relativeLayout.addView(flipper);
            relativeLayout.addView(levelupMessage.getLayout());
            relativeLayout.addView(taskCompletedMessage.getLayout());
            relativeLayout.addView(tutorialView.getLayout());
            setContentView(relativeLayout);
        } catch (Exception e) {
            onException(e);
        }
    }

    public void exitGame() {
        super.finish();
    }

    public boolean isShortMenuInGameActive() {
        return gameScreen.getGameViewManager().isShortMenuVisible();
    }

    public void flipToScreen(final Screen.ScreenType screen, final Screen.ScreenType caller) {
        if (screen != Screen.ScreenType.NONE && flipper.getCurrentScreen().getType() != screen) {
            runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    flipper.showScreen(screen, caller);
                }
            });
        }
    }

    public void flipToScreen(final Screen.ScreenType screen) {
        if (screen != Screen.ScreenType.NONE && flipper.getCurrentScreen().getType() != screen) {
            runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    flipper.showScreen(screen, flipper.getCurrentScreen().getType());
                }
            });
        }
    }

    public void flipToScreenNonUiThreaded(final Screen.ScreenType screen) {
        if (screen != Screen.ScreenType.NONE && flipper.getCurrentScreen().getType() != screen) {
            flipper.showScreen(screen, flipper.getCurrentScreen().getType());
        }
    }

    public void onSurfaceChanged() {
        if (flipper.getCurrentScreen().getType() == Screen.ScreenType.PRE_START)
            flipToScreen(Screen.ScreenType.START);
    }

    public void checkForAlreadyUnlockedPlayerShapes() {
        alreadyUnlockedPlayerShapes.clear();
        for (PlayerShape shape : PlayerShape.values()) {
            Task task = shape.getUnlockTask();
            task.update(levelManager);
            if (task.isConditionFulfilled())
                alreadyUnlockedPlayerShapes.add(shape);
        }
    }

    public void showNewFinishedPlayerShapeTasks() {
        List<PlayerShape> newUnlockedShapes = new ArrayList<>();
        for (PlayerShape shape : PlayerShape.values()) {
            Task task = shape.getUnlockTask();
            task.update(levelManager);
            if (task.isConditionFulfilled() && !alreadyUnlockedPlayerShapes.contains(shape)) {
                alreadyUnlockedPlayerShapes.add(shape);
                newUnlockedShapes.add(shape);
            }
        }
        if (!newUnlockedShapes.isEmpty())
            onTasksCompleted(newUnlockedShapes);
    }

    public void onWorldSelected(LevelPack levelPack) {
        levelsScreen.onLevelPackSelected(levelPack);
    }

    public void startGame(final LevelPack levelPack, final Level level) {
        flipToScreen(Screen.ScreenType.GAME);
        gameScreen.startGame(levelPack, level);
    }

    @Override
    public void onLvUp(final int level, boolean showLvUpMessage) {
        dataStorageHandler.writeUserData(user);
        if (showLvUpMessage) {
            runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    levelupMessage.show(level);
                }
            });
            tutorialManager.onLvUp(user);
        }
    }

    public void onTasksCompleted(final List<PlayerShape> shapes) {
        runOnUiThread(new Runnable() {
            @Override
            public void run() {
                taskCompletedMessage.show(shapes);
            }
        });
    }

    public void showTutorial(final List<BreakPoint> breakPoints) {
        if (!gameScreen.isLevelFinished()) {
            gameScreen.setGamePausedWithoutMenu();
        }
        runOnUiThread(new Runnable() {
            @Override
            public void run() {
                tutorialView.show(breakPoints);
            }
        });
    }

    public void onTutorialViewHidden() {
        if (!gameScreen.isLevelFinished())
            gameScreen.onResume();
    }

    public void resetData() {
        user.clearData();
        dataStorageHandler.writeUserData(user);
        dataStorageHandler.writeToolShopTutorialFinished(false);
        levelManager.reset();
        tutorialManager.resetAllTutorials();
        MyDatabase database = dataStorageHandler.getDatabase();
        database.open();
        database.clearLevelProgess();
        database.clearLevelPackLocked();
        for (ToolType type : ToolType.values())
            type.reset();
        database.writeToolData();
        database.close();
        checkForAlreadyUnlockedPlayerShapes();
    }

    @Override
    public boolean onKeyDown(int keyCode, KeyEvent event) {
        try {
            if (keyCode == KeyEvent.KEYCODE_BACK) {
                if (tutorialView.isShowingTutorial())
                    tutorialView.onClick(null);
                else
                    flipper.getCurrentScreen().onBackKeyDown();
                return true;
            }
        } catch (Exception e) {
            onException(e);
        }
        return super.onKeyDown(keyCode, event);
    }

    @Override
    public void onException(Exception e) {
        GameLog.e(e);
        super.finish();
    }

    @Override
    protected void onPause() {
        GameLog.d("OnPause");
        glSurfaceView.onPause();
        if (flipper.getCurrentScreen() == gameScreen)
            gameScreen.onPause();
        soundManager.activityPause();
        super.onPause();
    }

    @Override
    protected void onResume() {
        GameLog.d("OnResume");
        glSurfaceView.onResume();
        soundManager.activityResume();
        super.onResume();
    }

    @Override
    protected void onDestroy() {
        GameLog.d("OnDestroy");
        soundManager.destroy();

        super.onDestroy();
    }

    private boolean hasGLES20() {
        ActivityManager am = (ActivityManager) getSystemService(Context.ACTIVITY_SERVICE);
        ConfigurationInfo info = am.getDeviceConfigurationInfo();
        return info.reqGlEsVersion >= 0x20000;
    }

    public User getUser() {
        return user;
    }

    public DataStorageHandler getDataStorageHandler() {
        return dataStorageHandler;
    }

    public SoundManager getSoundManager() {
        return soundManager;
    }

    public Typeface getTypeface() {
        return typeface;
    }

    public LevelManager getLevelManager() {
        return levelManager;
    }

    public TutorialManager getTutorialManager() {
        return tutorialManager;
    }
}
